// -*- Mode: Java -*-

/*--------------------------- BEGIN LICENSE BLOCK ---------------------------+
|                                                                            |
| Version: MPL 1.1/GPL 2.0/LGPL 2.1                                          |
|                                                                            |
| The contents of this file are subject to the Mozilla Public License        |
| Version 1.1 (the "License"); you may not use this file except in           |
| compliance with the License. You may obtain a copy of the License at       |
| http://www.mozilla.org/MPL/                                                |
|                                                                            |
| Software distributed under the License is distributed on an "AS IS" basis, |
| WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License   |
| for the specific language governing rights and limitations under the       |
| License.                                                                   |
|                                                                            |
| The Original Code is the Knowledge Resolver System.                        |
|                                                                            |
| The Initial Developer of the Original Code is                              |
| UNIVERSITY OF SOUTHERN CALIFORNIA, INFORMATION SCIENCES INSTITUTE          |
| 4676 Admiralty Way, Marina Del Rey, California 90292, U.S.A.               |
|                                                                            |
| Portions created by the Initial Developer are Copyright (C) 2010-2015      |
| the Initial Developer. All Rights Reserved.                                |
|                                                                            |
| Contributor(s):                                                            |
|                                                                            |
| Alternatively, the contents of this file may be used under the terms of    |
| either the GNU General Public License Version 2 or later (the "GPL"), or   |
| the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),   |
| in which case the provisions of the GPL or the LGPL are applicable instead |
| of those above. If you wish to allow use of your version of this file only |
| under the terms of either the GPL or the LGPL, and not to allow others to  |
| use your version of this file under the terms of the MPL, indicate your    |
| decision by deleting the provisions above and replace them with the notice |
| and other provisions required by the GPL or the LGPL. If you do not delete |
| the provisions above, a recipient may use your version of this file under  |
| the terms of any one of the MPL, the GPL or the LGPL.                      |
|                                                                            |
+---------------------------- END LICENSE BLOCK ----------------------------*/

// Version: $Id$
//
// Author: Hans Chalupsky


// Toolkit to index, host, query and search Freebase and its variants (e.g., BaseKB) as a Lucene index.
// See the README file for more information.

// Libraries we need:
// - lucene-5.2.1/core/lucene-core-5.2.1.jar
// - lucene-5.2.1/analysis/common/lucene-analyzers-common-5.2.1.jar
// - lucene-5.2.1/queryparser/lucene-queryparser-5.2.1.jar
// - args4j/2.0.23/args4j-2.0.23.jar
//
// compile like this (assumes the above libraries in `lib'):
// % javac -cp '.:bin:lib/*' -d bin FreebaseTools.java
//
// run like this (more examples in the README):
// % java -cp '.:bin:lib/*' edu.isi.kres.FreebaseTools -T data/freebase-rdf-latest.shrink.sort.gz  -I data/freebase-rdf-latest.shrink.sort.index  -c index -f -v      // 56min, 13GB index
// % java -cp '.:bin:lib/*' edu.isi.kres.FreebaseTools -T data/freebase-rdf-latest.shrink.sort.gz  -I data/freebase-rdf-latest.shrink.sort.index  -c index -f -o -v   // 71min, 13GB index
// % java -cp '.:bin:lib/*' edu.isi.kres.FreebaseTools -T data/basekb-gold-jan-2015.shrink.sort.gz -I data/basekb-gold-jan-2015.shrink.sort.index -c index -f -o -v   // 51min, 11GB index
// % java -cp '.:bin:lib/*' edu.isi.kres.FreebaseTools -I data/basekb-gold-jan-2015.shrink.sort.index -c lookup -q f_m.0h54qv8 -v
//
// these search commands assume an existing index properly configured in the shell scripts:
// % ./fbt-lookup.sh -q f_m.0h54qv8 -v
// % ./fbt-search.sh -q 'Claude AND Parsons AND r_type:f_people.person'
// % ./fbt-search.sh -q 'Parsons AND r_type:f_people.person' -p subject -v
//
// Useful info about Lucene query syntax:
// - https://lucene.apache.org/core/2_9_4/queryparsersyntax.html


package edu.isi.kres;

// Java:
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.nio.charset.StandardCharsets;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.zip.GZIPInputStream;

// Args4J:
import org.kohsuke.args4j.CmdLineException;
import org.kohsuke.args4j.CmdLineParser;
import org.kohsuke.args4j.Option;

// Lucene:
import org.apache.lucene.analysis.Analyzer;
import org.apache.lucene.analysis.standard.StandardAnalyzer;
import org.apache.lucene.codecs.lucene50.Lucene50Codec;
import org.apache.lucene.codecs.lucene50.Lucene50StoredFieldsFormat.*;
import org.apache.lucene.document.Document;
import org.apache.lucene.document.DocumentStoredFieldVisitor;
import org.apache.lucene.document.Field;
import org.apache.lucene.document.StoredField;
import org.apache.lucene.document.StringField;
import org.apache.lucene.document.TextField;
import org.apache.lucene.index.DirectoryReader;
import org.apache.lucene.index.IndexableField;
import org.apache.lucene.index.IndexReader;
import org.apache.lucene.index.IndexWriter;
import org.apache.lucene.index.IndexWriterConfig.OpenMode;
import org.apache.lucene.index.IndexWriterConfig;
import org.apache.lucene.index.Term;
import org.apache.lucene.queryparser.classic.QueryParser;
import org.apache.lucene.search.IndexSearcher;
import org.apache.lucene.search.Query;
import org.apache.lucene.search.ScoreDoc;
import org.apache.lucene.search.TermQuery;
import org.apache.lucene.search.TopDocs;
import org.apache.lucene.store.Directory;
import org.apache.lucene.store.FSDirectory;


public class FreebaseTools {

       //
    ////// Command line and configuration options
     //

    @Option(name = "-c", aliases = { "--command" }, required = false,
            usage = "Command to execute, one of `lookup' (the default), `search' or `index'")
    String command = "lookup";

    @Option(name = "-T", aliases = { "--triples" }, required = false,
            usage = "Compressed, abbreviated N-triples KB file")
    String triplesFile = "";

    @Option(name = "-I", aliases = { "--index" }, required = false,
            usage = "Lucene index directory")
    String indexDirectoryName = "";

    @Option(name = "-P", aliases = { "--indexed-preds" }, required = false,
            usage = "Predicates file listing predicates that should be indexed so they can be restricted in queries")
    String indexedPredsFile = "indexed-preds.lst";

    @Option(name = "-o", aliases = { "--optimize-index" }, required = false,
            usage = "Optimize index by merging it onto a single segment for faster startup and search (can take a while)")
    boolean optimizeIndex = false;

    @Option(name = "-f", aliases = { "--full-data" }, required = false,
            usage = "Index the full data, otherwise only a small test portion will be indexed")
    boolean indexFullData = false;

    @Option(name = "-q", aliases = { "--query" }, required = false,
            usage = "Query string for lookup and search queries.  Accepts full Lucene query syntax for `search' queries.  `lookup' queries take a comma/white-space separated list of subject keys, if `-' is given, subjects will be read from stdin.")
    String queryString = "supercalifragilisticexpialidocious"; // 161 matching subjects

    @Option(name = "-m", aliases = { "--max-hits" }, required = false,
            usage = "Maximum number of `search' query results to generate, default is 10, use -1 to get all results")
    int maxHits = 10;

    @Option(name = "-p", aliases = { "--print-mode" }, required = false,
            usage = "Result print mode, one of `all' (default), `subject', or a comma/white-space separated list of predicates/chains (e.g., 'p1, p2>p3, p4')")
    String printMode = "all";

    @Option(name = "-v", aliases = { "--show-progress" }, required = false,
            usage = "Show progress and timing information")
    boolean showProgress = false;

    @Option(name = "-d", aliases = { "--show-debug" }, required = false,
            usage = "Show debugging information")
    boolean showDebug = false;

    @Option(name = "-h", aliases = { "--help" }, required = false,
            usage = "Show usage information")
    boolean showUsage = false;

    // If true, predicate chains will not loop back to their seeds:
    boolean suppressPredicateLoops = true;

    // Default field used by search queries if a field is not explicitly specified:
    String defaultSearchField = normalizeUri("<rs_label>");


       //
    ////// Top-level Command Dispatch
     //

    public static void main(String[] args) {
        FreebaseTools tools = new FreebaseTools();
        CmdLineParser parser = new CmdLineParser(tools);
        try {
            parser.parseArgument(args);
        }
        catch (CmdLineException e) {
            System.err.println("ERROR: " + e.getMessage());
            System.err.println("Usage:");
            parser.printUsage(System.err);
            System.exit(0);
        }

        if (tools.showUsage) {
            System.err.println("Usage:");
            parser.printUsage(System.err);
        }
        else if (tools.command.equals("index")) {
            tools.doIndexTriples();
        }
        else if (tools.command.equals("lookup")) {
            tools.doLookup();
        }
        else if (tools.command.equals("search")) {
            tools.doSearch();
        }
        //else if (tools.command.equals("read-triples")) {
        //    tools.doReadCompressedTriples();
        //}
        else {
            System.err.println("Usage:");
            parser.printUsage(System.err);
        }
        System.exit(0);
    }


       //
    ////// Utilities:
     //

    public BufferedReader openTextFile(String filename) throws IOException {
        // Open a UTF8-encoded text input stream to the (potentially compressed) file `filename'.
        InputStream in = new FileInputStream(filename);
        if (filename.endsWith(".gz"))
            in = new GZIPInputStream(in);
        return new BufferedReader(new InputStreamReader(in, StandardCharsets.UTF_8));
    }

    public List<String> readList(BufferedReader in, long maxLines) throws IOException {
        // Read non-empty, non-commented lines from `in' and return the result in a list.
        // If `maxLines' is >= 0, remove at most that many items (supply -1 for unlimited lines).
        String line = null;
        List<String> list = new ArrayList<String>();
        
        while (maxLines != 0) {
            line = in.readLine();
            if (line == null)
                break;
            if ((line.startsWith("#")) || (line.matches("^\\s*$")))
                // skip comments and blank lines:
                continue;
            else {
                list.add(line);
                maxLines--;
            }
        }
        return list;
    }

    public List<String> readListFile(String filename) throws IOException {
        BufferedReader in = openTextFile(filename);
        List<String> list = readList(in, -1);
        in.close();
        return list;
    }

    public void showProgress(long count, int defaultInterval) {
        if (showProgress && ((count % defaultInterval) == 0)) {
            System.err.print(".");
            System.err.flush();
        }
    }


       //
    ////// Index Creation and Access
     //

    // Lucene compression info:
    // - http://blog.florian-hopf.de/2012/12/looking-at-plaintext-lucene-index.html
    // - https://lucene.apache.org/core/4_4_0/core/org/apache/lucene/codecs/compressing/CompressingStoredFieldsFormat.html
    // - https://lucene.apache.org/core/4_4_0/core/org/apache/lucene/codecs/compressing/package-summary.html
    // It looks like the default storage format compresses now, so we don't have to do anything special.

    public String getIndexDirectoryName() throws IOException {
        // Either return the provided directory or derive one based on `triplesFile'.
        if (! indexDirectoryName.equals(""))
            return indexDirectoryName;
        else if (! triplesFile.equals(""))
            return triplesFile + ".lucene.index";
        else
            throw new IOException("ERROR: Cannot determine Lucene index directory, supply via -i");
    }

    // We assume we either are creating a new index or querying it, but never both.
    // These variables cache various index-related objects for either reading or writing:
    Directory indexDirectory = null;
    IndexWriter indexWriter = null;
    IndexReader indexReader = null;
    IndexSearcher indexSearcher = null;
    Analyzer indexAnalyzer = null;

    public void initializeIndex() throws IOException {
        // Create a new index directory and writer to index a triples file.
        // Raise an error if an index already exists, so we don't accidentally overwrite it.
        String indexDir = getIndexDirectoryName();
        if ((new java.io.File(indexDir)).isDirectory())
            throw new IOException("ERROR: Index directory already exists, remove it before indexing");

        indexDirectory = FSDirectory.open(Paths.get(indexDir));
        indexAnalyzer = new StandardAnalyzer();
        IndexWriterConfig iwc = new IndexWriterConfig(indexAnalyzer);

        // we always create a new index from scratch:
        iwc.setOpenMode(OpenMode.CREATE);
        iwc.setCodec(new Lucene50Codec(Mode.BEST_SPEED));          // the default
        //iwc.setCodec(new Lucene50Codec(Mode.BEST_COMPRESSION));  // slower but better compression

        indexWriter = new IndexWriter(indexDirectory, iwc);
    }

    public void finalizeIndex() throws IOException {
        // To maximize search performance, we call forceMerge here (slow), but since we
        // are dealing with a totally static index, the cost should be worth it:
        if (optimizeIndex) {
            if (showProgress) System.err.println("Optimizing index...");
            indexWriter.forceMerge(1);
            indexWriter.commit();
        }
        indexWriter.close();
        indexDirectory = null;
        indexAnalyzer = null;
        indexWriter = null;
    }

    public void initializeSearch() throws IOException {
        // Initialize index structures for search.
        if (indexDirectory == null)
            indexDirectory = FSDirectory.open(Paths.get(getIndexDirectoryName()));
    }

    public IndexReader getIndexReader() throws IOException {
        // Get the current index reader or create it.
        if (indexReader == null) {
            initializeSearch();
            indexReader = DirectoryReader.open(indexDirectory);
        }
        return indexReader;
    }

    public IndexSearcher getIndexSearcher() throws IOException {
        // Get the current index searcher or create it.
        if (indexSearcher == null)
            indexSearcher = new IndexSearcher(getIndexReader());
        return indexSearcher;
    }

    public Analyzer getIndexAnalyzer() throws IOException {
        // Get the current index analyzer or create it.
        if (indexAnalyzer == null)
            indexAnalyzer = new StandardAnalyzer();
        return indexAnalyzer;
    }

    public long getIndexNumDocs() throws IOException {
        // Returns the number of documents/subjects indexed in the index.
        return getIndexReader().numDocs();
    }

    public long getIndexTermNumDocs(Term term) throws IOException {
        // Return the number of documents/subjects containing `term'.
        return getIndexReader().docFreq(term);
    }


       //
    ////// Indexing Triples
     //

    Set<String> indexedPreds = null;

    void initializeIndexedPreds() throws IOException {
        indexedPreds = new HashSet<String>();
        for (String pred : readListFile(indexedPredsFile)) {
            indexedPreds.add(normalizeUri(pred));
        }
    }

    public boolean isIndexedPredicate(String predicate) throws IOException {
        // Return true if `predicate' should be indexed in addition to just being stored.
        // `predicate' is expected to be a normalized URI.
        if (indexedPreds == null)
            initializeIndexedPreds();
        return indexedPreds.contains(predicate);
    }

    public void indexTriples() throws IOException {
        // Index the triples in `triplesFile' to a Lucene index.
        // This assumes that `triplesFile' is sorted by subject.
        // For each subject we collect all its predicate/value pairs and index them
        // as a single Lucene document.  We create a subject field which will serve
        // as a primary key for direct lookup.  Each predicate becomes the name of a field,
        // if there are multiple values they become multiple fields with the same name.
        // All predicate/value pairs are also stored, so we can retrieve the info
        // directly from the index without having to go to a separate database.
        // Lucene stored field compression will provide good compression and access speed.

        BufferedReader in = openTextFile(triplesFile);

        String line = null;
        String[] triple;
        String subject = "";
        long count = 0;
        long maxTriples = indexFullData ? Long.MAX_VALUE : 20000000;

        // this map collects values for each of a subject's predicates -
        // we use a TreeMap for per-subject predicate sorting:
        Map<String,List<String>> predValues = new TreeMap<String,List<String>>();
        
        if (showProgress) System.err.println("Indexing to directory '" + getIndexDirectoryName() + "'...");
        while (true) {
            line = in.readLine();
            if ((line == null) || (count >= maxTriples)) {
                indexRecord(indexWriter, subject, predValues);
                break;
            }

            triple = line.split("\t");
            if (! triple[0].equals(subject)) {
                if (! "".equals(subject))
                    // new subject, index the current subject and start a new one:
                    indexRecord(indexWriter, subject, predValues);
                subject = triple[0];
                predValues.clear();
            }

            // record this predicate and value for the current subject:
            String predicate = triple[1];
            String value = triple[2];
            List<String> values = predValues.get(predicate);
            if (values == null) {
                values = new ArrayList<String>(5);
                predValues.put(predicate, values);
            }
            values.add(value);
            ++count;
            showProgress(count, 1000000);
        }
        if (showProgress) System.err.println();
    }

    public boolean isUri(String x) {
        // Return true if `x' is a URI.
        return x.startsWith("<");
    }

    public String normalizeUri(String uri) {
        // Do some URI normalization to make things go more smoothly with the query parser.
        // We want lower-case namespace IDs and no `:'s, since those make the query field parser unhappy.
        // That way we can freely mix text queries with exact field restrictions such as type names.
        // This is now done mostly by the preprocessing steps, and we only have to strip off angle brackets.
        if (uri.charAt(0) == '<')
            return uri.substring(1, uri.length()-1);
        else
            return uri;
    }

    public void indexRecord(IndexWriter writer, String subject, Map<String,List<String>> predValues) throws IOException {
        Document doc = new Document();
        Field subjField = new StringField("subject", normalizeUri(subject), Field.Store.YES);
        //System.err.println("subject: " + subject);
        doc.add(subjField);

        for (Map.Entry<String, List<String>> entry : predValues.entrySet()) {
            String predicate = normalizeUri(entry.getKey());
            List<String> values = entry.getValue();

            //System.err.println(predicate + ": " + values);
            for (String value : values) {
                boolean isUriValue = isUri(value);
                if (isUriValue)
                    value = normalizeUri(value);
                if (isIndexedPredicate(predicate)) {
                    if (isUriValue)
                        // treat URI elements as atomic strings (e.g., types):
                        doc.add(new StringField(predicate, value, Field.Store.YES));
                    else
                        // all others, run through the analyzer - this might need refinement:
                        doc.add(new TextField(predicate, value, Field.Store.YES));
                }
                else
                    doc.add(new StoredField(predicate, value));
            }
        }

        // we are creating the index from scratch, so we just add the document:
        writer.addDocument(doc);
    }


       //
    ////// Subject/Predicate API
     //

    public int getSubjectDocID(String subjectURI) throws IOException {
        // Retrieve the document ID identified by `subjectURI', return -1 if nothing was found.
        // Not sure if a TermQuery is the fastest way to do a primary key lookup in Lucene.
        TermQuery query = new TermQuery(new Term("subject", subjectURI));
        TopDocs result = getIndexSearcher().search(query, 1);
        if (result.totalHits == 0)
            return -1;
        else {
            return result.scoreDocs[0].doc;
        }
    }

    public Document getSubjectDoc(String subjectURI) throws IOException {
        // Retrieve the document identified by `subjectURI', return null if nothing was found.
        int subjectId = getSubjectDocID(subjectURI);
        return (subjectId < 0) ? null : getIndexReader().document(subjectId);
    }

    public String getSubjectName(Document subjectDoc) throws IOException {
        // Return the string name of `subjectDoc'.
        return subjectDoc.get("subject");
    }

    public String getSubjectPredicateValue(Document subjectDoc, String predName) throws IOException {
        // Return the value of predicate `predName' on `subjectDoc'.  If there are muliple values,
        // return the first one indexed, if there are none, return null.
        return subjectDoc.get(predName);
    }

    public String[] getSubjectPredicateValues(Document subjectDoc, String predName) throws IOException {
        // Return the values of predicate `predName' on `subjectDoc' in the order they were indexed.
        // If there are none, return an empty array.
        return subjectDoc.getValues(predName);
    }

    public String getSubjectPredicateValue(String subjectURI, String predName) throws IOException {
        // Return the value of predicate `predName' on `subjectURI'.  If there are muliple values,
        // return the first one indexed, if there are none, return null.
        // This is specialized to only retrieve the `predName' field of the subject document.
        // If the full document has already been retrieved, use the Document accessor instead.
        int subjectId = getSubjectDocID(subjectURI);
        if (subjectId < 0)
            return null;
        else {
            DocumentStoredFieldVisitor fieldVisitor = new DocumentStoredFieldVisitor(predName);
            getIndexReader().document(subjectId, fieldVisitor);
            Document subject = fieldVisitor.getDocument();
            return getSubjectPredicateValue(subject, predName);
        }
    }

    String[] emptyValues = new String[0];

    public String[] getSubjectPredicateValues(String subjectURI, String predName) throws IOException {
        // Return the values of predicate `predName' on `subjectURI'.
        // If there are none, return an empty array.
        // This is specialized to only retrieve the `predName' field of the subject document.
        // If the full document has already been retrieved, use the Document accessor instead.
        int subjectId = getSubjectDocID(subjectURI);
        if (subjectId < 0)
            return emptyValues;
        else {
            DocumentStoredFieldVisitor fieldVisitor = new DocumentStoredFieldVisitor(predName);
            getIndexReader().document(subjectId, fieldVisitor);
            Document subject = fieldVisitor.getDocument();
            return getSubjectPredicateValues(subject, predName);
        }
    }


       //
    ////// Printing Results
     //

    public void printSubject(Document subject, float score, PrintWriter out) throws IOException {
        // Print `subject''s name to `out', also print its `score' if it is non-negative.
        if (score >= 0.0) {
            out.print(getSubjectName(subject));
            out.print("\t");
            out.println(score);
        }
        else
            out.println(getSubjectName(subject));
    }

    public void printSubjectAllPredicates(Document subject, float score, PrintWriter out) throws IOException {
        // Pretty-print everything we know about `subject' to `out'.
        // Annotate its `score' if it is non-negative.
        if (score >= 0.0)
            out.println(getSubjectName(subject) + ": [score=" + score + "]");
        else
            out.println(getSubjectName(subject) + ":");
        for (IndexableField field : subject.getFields()) {
            if (! "subject".equals(field.name()))
                out.println("    " + field.name() + ": " + field.stringValue());
        }
    }

    // Printing predicate chains:

    List<String> valueRow = new ArrayList<String>();

    public List<String> getValueRow() {
        // Use a shared value row buffer for now.
        return valueRow;
    }

    public void pushValueRow(List<String> row, String value) {
        row.add(value);
    }

    public void popValueRow(List<String> row) {
        row.remove(row.size()-1);
    }

    public void printValueRow(List<String> row, float score, String[] predicateChain, PrintWriter out) {
        out.print(row.get(0));
        if (score >= 0.0) {
            out.print("\t");
            out.print(score);
        }
        int nValues = row.size();
        for (int i = 1; i < nValues; i++) {
            out.print("\t");
            out.print(predicateChain[i-1]);
            out.print("\t");
            out.print(row.get(i));
        }
        out.println();
    }

    public void printPredicateChain(String subjectUri, float score, String[] predicateChain, int start, List<String> row, PrintWriter out) throws IOException {
        // Helper for `printSubjectPredicates' to print the tail of a predicate chain.
        if (suppressPredicateLoops && row.contains(subjectUri))
            // don't print value chains that lead back to an earlier seed value:
            return;
        pushValueRow(row, subjectUri);
        if (start >= predicateChain.length)
            printValueRow(row, score, predicateChain, out);
        else {
            String pred = predicateChain[start];
            String[] values = getSubjectPredicateValues(subjectUri, pred);
            if (values.length > 0) {
                for (String value : values) {
                    printPredicateChain(value, score, predicateChain, start+1, row, out);
                }
            }
            else {
                pushValueRow(row, "null");
                printValueRow(row, score, predicateChain, out);
                popValueRow(row);
            }
        }
        popValueRow(row);
    }

    public void printSubjectPredicates(Document subject, float score, String[][] predicates, PrintWriter out) throws IOException {
        // Print all of `subject's `predicates' that have values to `out'.
        // Handles predicate value chains if necessary.
        // Also print `score' after `subject' if it is non-negative.

        String subjectName = getSubjectName(subject);
        List<String> row = getValueRow();
        pushValueRow(row, subjectName);
        for (String[] chain : predicates) {
            String pred = chain[0];
            String[] values = getSubjectPredicateValues(subject, pred);
            
            if (values.length > 0) {
                for (String value : values)
                    printPredicateChain(value, score, chain, 1, row, out);
            }
            else {
                pushValueRow(row, "null");
                printValueRow(row, score, chain, out);
                popValueRow(row);
            }
        }
        popValueRow(row);
    }


    public String[] parsePredicateList(String predicateList) {
        // Parse a white-space or comma-separated `predicateList' into its components.
        return predicateList.trim().split("(\\s|,)+");
    }

    public String[] parsePredicateChain(String predicateChain) {
        // Parse a >-separated `predicateChain' into its components.
        // Since `>' is a URI-delimiter, it won't show up in predicate names.
        return predicateChain.split(">");
    }

    String[][] printModePredicates = null;    // list of predicate chains

    public String[][] getPrintModePredicates() {
        if (printModePredicates == null) {
            String[] preds = parsePredicateList(printMode);
            printModePredicates = new String[preds.length][];
            for (int i = 0; i < preds.length; i++)
                printModePredicates[i] = parsePredicateChain(preds[i]);
        }
        return printModePredicates;
    }

    HashSet<String> printModeFields = null;  // list of fields to restrict Document access to, only used for top-level queries

    public Set<String> getPrintModeFields() {
        if ((printModeFields == null) && "all".equals(printMode))
            return null;
        else if (printModeFields == null) {
            printModeFields = new HashSet<String>(Arrays.asList("subject"));
            for (String[] pred : getPrintModePredicates())
                printModeFields.add(pred[0]);
        }
        return printModeFields;
    }

    public void printSubjectInMode(Document subject, float score, String mode, PrintWriter out) throws IOException {
        // Print `subject' and `score' onto `out' according to `mode'.
        if ("subject".equals(mode))
            printSubject(subject, score, out);
        else if ("all".equals(mode))
            printSubjectAllPredicates(subject, score, out);
        else
            printSubjectPredicates(subject, score, getPrintModePredicates(), out);
    }

    public Document getDocumentInMode(int docid) throws IOException {
        // Access a Document from its `docid' according to the current print mode.
        // This ensures we are materializing the minimal number of document fields necessary.
        Set<String> displayFields = getPrintModeFields();
        if (displayFields == null)
            return getIndexReader().document(docid);
        else
            return getIndexReader().document(docid, displayFields);
    }


       //
    ////// Command Implementations
     //

    public void doIndexTriples() {
        // Implements `index' command.
        try {
            if (! indexFullData)
                System.err.println("INFO: Running in test mode with 20M triples, use -f to index all the data");
            initializeIndex();
            indexTriples();
            finalizeIndex();
        }
        catch (IOException e) {
            System.err.println("ERROR: " + e.getMessage());
            System.exit(1);
        }
    }

    public void doLookup() {
        try {
            long timeSetup = 0;
            long timeQuery = 0;
            long timeDisplay = 0;

            // force index initialization, so we get a better sense of the actual query execution time:
            if (showProgress) System.err.println("Loading index...");
            long timeStart = java.lang.System.currentTimeMillis();
            getIndexReader();
            timeSetup += (java.lang.System.currentTimeMillis() - timeStart);

            int blockSize = 1000;
            BufferedReader in = queryString.equals("-") ? new BufferedReader(new InputStreamReader(System.in)) : null;
            List<String> subjects = (in != null) ? readList(in, blockSize) : Arrays.asList(parsePredicateList(queryString));
            PrintWriter out = new PrintWriter(new BufferedWriter(new OutputStreamWriter(System.out))); // Java I/O API is insane
            boolean fullDisplay = "all".equals(printMode);

            while (true) {
                for (String subject : subjects) {
                    // now run the query:
                    timeStart = java.lang.System.currentTimeMillis();
                    int docid = getSubjectDocID(subject);
                    if (docid < 0) {
                        timeQuery += (java.lang.System.currentTimeMillis() - timeStart);
                        System.err.println("Not found: " + subject);
                    }
                    else {
                        Document doc = getDocumentInMode(docid);
                        timeQuery += (java.lang.System.currentTimeMillis() - timeStart);
                        timeStart = java.lang.System.currentTimeMillis();
                        printSubjectInMode(doc, -1, printMode, out);
                        if (fullDisplay)
                            out.println();
                        timeDisplay += (java.lang.System.currentTimeMillis() - timeStart);
                    }
                }
                if (in == null)
                    break;
                subjects = readList(in, blockSize);
                if (subjects.isEmpty())
                    break;
            }
            out.flush();
            if (showProgress) 
                System.err.println("Run time: setup=" + timeSetup + "ms, query=" + timeQuery + "ms, display=" + timeDisplay + "ms");
            getIndexReader().close();
        }
        catch (Exception e) {
            System.err.println("ERROR: " + e.getMessage());
            System.exit(1);
        }
    }

    public void doSearch() {
        try {
            long timeSetup = 0;
            long timeQuery = 0;
            long timeDisplay = 0;

            // force index initialization, so we get a better sense of the actual query execution time:
            if (showProgress) System.err.println("Loading index...");
            long timeStart = java.lang.System.currentTimeMillis();
            getIndexSearcher();
            getIndexAnalyzer();
            timeSetup += (java.lang.System.currentTimeMillis() - timeStart);
            if (showProgress) System.err.println("Index contains " + getIndexNumDocs() + " documents");

            timeStart = java.lang.System.currentTimeMillis();
            Query query = new QueryParser(defaultSearchField, getIndexAnalyzer()).parse(queryString);
            if (showProgress) System.err.println("Searching for: " + query.toString(defaultSearchField));
            boolean allResults = (maxHits == -1);
            int nDocs = allResults ? 1 : maxHits;
            TopDocs results = getIndexSearcher().search(query, nDocs);
            ScoreDoc[] hits = results.scoreDocs;
    
            int numTotalHits = results.totalHits;
            if (showProgress) System.err.println("Found " + numTotalHits + " matching subject(s)");
            if (allResults) {
                nDocs = numTotalHits;
                if (showProgress) System.err.println("Collecting results...");
                results = getIndexSearcher().search(query, numTotalHits);
                hits = results.scoreDocs;
            }
            timeQuery = java.lang.System.currentTimeMillis() - timeStart;
      
            timeStart = java.lang.System.currentTimeMillis();
            if (showProgress) System.err.println("Printing results...");
            PrintWriter out = new PrintWriter(new BufferedWriter(new OutputStreamWriter(System.out))); // Java I/O API is insane
            int end = Math.min(hits.length, nDocs);
            boolean fullDisplay = "all".equals(printMode);

            for (int i = 0; i < end; i++) {
                // this takes about 90s for 3M subjects names, so not super fast but ok (0.03ms per subject)
                int docid = hits[i].doc;
                float score = allResults ? -1.0f : hits[i].score;
                Document doc = getDocumentInMode(docid);
                printSubjectInMode(doc, score, printMode, out);
                if (showDebug)
                    out.print("Score Explanation: " + getIndexSearcher().explain(query, docid).toString());
                if (fullDisplay)
                    out.println();
            }
            out.flush();
            timeDisplay = java.lang.System.currentTimeMillis() - timeStart;
            if (showProgress) 
                System.err.println("Run time: setup=" + timeSetup + "ms, query=" + timeQuery + "ms, display=" + timeDisplay + "ms");
            getIndexReader().close();
        }
        catch (Exception e) {
            System.err.println("ERROR: " + e.getMessage());
            System.exit(1);
        }
    }


    public void readCompressedTriples () throws IOException {
        // Test driver to check reading performance: this reads and counts 878M lines in 7m:40s which is
        // fairly comparable to the 5m:40s it takes with `zcat <file> | wc'.
        BufferedReader in = openTextFile(triplesFile);
        String line = null;
        long count = 0;
        
        while (true) {
            line = in.readLine();
            if (line == null)
                break;
            ++count;
            showProgress(count, 1000000);
        }
        System.err.println();
        System.err.println("total count=" + count);
        in.close();
    }

    public void doReadCompressedTriples() {
        try {
            readCompressedTriples();
        }
        catch (Exception e) {
            System.err.println("ERROR: " + e.getMessage());
            System.exit(1);
        }
    }
}
