#!/bin/csh -f

# Utilitity to shrink a Freebase or BaseKB RDF n-triples archive into something more managable.
#
# Usage: fbt-shrink-freebase.sh [-f] [-o <output-file>] <freebase-file>...
#
# <freebase-file> the file(s) containing the Freebase data dump which can be full-text or compressed;
#                 if a list of files is given, they all need to use the same format and compression.
# <output-file>   names the file to write the result to, should have a .gz suffix.
#                 If none is supplied, the output file will be derived from the first <freebase-file>.
# -f              if supplied, the full <freebase-file>s will be processed; otherwise, we are
#                 running in test mode where only 1,000,000 triples are processed for testing.


set SCRIPT=`basename $0`
set FBT_HOME=`dirname $0`
set FBT_HOME=`readlink -f $FBT_HOME`

set input_files=
set output_file=
set test_mode=1

# see /usr/share/getopt/getopt-parse.tcsh:
set parsed_opt=(`getopt -s tcsh -o fo: -- $argv:q`)
if ($? != 0) then 
  echo "$SCRIPT : illegal options" >/dev/stderr
  exit 1
endif

eval set argv=\($parsed_opt:q\)

while (1)
	switch($1:q)
	case -f:
		set test_mode=0; shift
		breaksw;
	case -o:
		set output_file=$2:q ; shift ; shift
		breaksw;
	case --:
		shift
                set input_files=($argv)
		break
	default:
		echo "$SCRIPT : internal error" ; exit 1
	endsw
end

if ($#input_files == 0) then
    echo $SCRIPT : missing "<freebase-file>"
    exit 1
endif

if ($output_file == '') then
    set output_file=`dirname $input_files[1]`/`basename $input_files[1] .gz`.shrink.sort.gz
endif

# make them absolute pathnames:
set ni=$#input_files
while ($ni >= 1)
    set input_files[$ni]=`readlink -f $input_files[$ni]`
    @ ni--
end
set output_file=`readlink -f $output_file`

if (-e $output_file) then
    echo "$SCRIPT : output file $output_file already exists, remove it first"
    exit 1
endif

# for the unlikely case that somebody uses an uncompressed triples archive:
if ( `file $input_files[1] | grep -c 'gzip compressed'` ) then
   set CAT='zcat'
else
   set CAT='cat'
endif

if (`$CAT $input_files | head -1000 | grep -c '^<http://rdf.freebase.com/'` > 0) then
    echo "$SCRIPT : detected Freebase data format"
    set FBT_ABBREV=./fbt-abbrev-freebase-uris.sh
else if (`$CAT $input_files | head -1000 | grep -c '^<http://rdf.basekb.com/'` > 0) then
    echo "$SCRIPT : detected BaseKB data format"
    set FBT_ABBREV=./fbt-abbrev-basekb-uris.sh
else
    echo "$SCRIPT : failed to determine if this is Freebase or BaseKB data"
    exit 1
endif


cd $FBT_HOME
echo "$SCRIPT : Shrinking $input_files[1]... to $output_file..."
if ($test_mode) then
    echo "$SCRIPT : Running on 1,000,000 triples only in test mode, use -f to run on the full set"
    $CAT $input_files | head -10000000 | tail -1000000 | \
         $FBT_ABBREV | ./fbt-filter-langs.sh | ./fbt-filter-preds.sh | ./fbt-filter-values.sh | ./fbt-sort-triples.sh | ./fbt-compress.sh > $output_file
else
    $CAT $input_files | $FBT_ABBREV | ./fbt-filter-langs.sh | ./fbt-filter-preds.sh | ./fbt-filter-values.sh | ./fbt-sort-triples.sh | ./fbt-compress.sh > $output_file
endif
